import logging
import os
import sys

import daemon
from daemon.pidfile import TimeoutPIDLockFile
import lockfile

from baltrad.bdbcommon import optparse
from baltrad.bdbserver import backend, config
from baltrad.bdbserver.web import app

logger = logging.getLogger("baltrad.bdbserver")

def excepthook(*exc_info):
    logger.error("unhandled exception", exc_info=exc_info)
    sys.exit(1)

def create_optparser():
    optparser = optparse.create_parser()
    optparser.set_usage(
        "%s [--conf=CONFFILE] [ARGS]" % (
            os.path.basename(sys.argv[0])
        )
    )
    optparser.add_option(
        "--conf", type="path", dest="conffile",
        help="configuration file",
    )
    return optparser

def create_backend(config):
    type_ = config.get("baltrad.bdb.server.backend.type")
    try:
        backend_cls = backend.Backend.get_impl(type_)
    except LookupError:
        raise LookupError(
            "unknown baltrad.bdb.server.backend.type: %s" % type_
        )
    return backend_cls.from_conf(config)

def read_config(conffile):
    if not conffile:
        raise SystemExit("configuration file not specified")
    try:
        return config.Properties.load(conffile)
    except IOError:
        raise SystemExit("failed to read configuration from " + conffile)
   
def run_create():
    optparser = create_optparser()

    opts, args = optparser.parse_args()
    config = read_config(opts.conffile)

    backend = create_backend(config)
    backend.create()

def run_drop():
    optparser = create_optparser()

    opts, args = optparser.parse_args()

    conf = read_config(opts.conffile)

    backend = create_backend(conf)
    backend.drop()

def run_upgrade():
    optparser = create_optparser()

    opts, args = optparser.parse_args()

    conf = read_config(opts.conffile)

    backend = create_backend(conf)
    backend.upgrade()

def get_logging_level(conf):
    v = conf.get("baltrad.bdb.server.log.level", "INFO")
    if v == "DEBUG":
        return logging.DEBUG
    elif v == "INFO":
        return logging.INFO
    elif v == "WARN":
        return logging.WARN
    elif v == "WARNING":
        return logging.WARNING
    elif v == "ERROR":
        return logging.ERROR
    else:
        return logging.INFO

def configure_logging(opts, level=logging.INFO):
    logger = logging.getLogger()
    logger.setLevel(level)

    if opts.foreground:
        logger.addHandler(logging.StreamHandler(sys.stdout))
    if opts.logfile:
        logger.addHandler(logging.FileHandler(opts.logfile))

## Checks if the process with provided pid is running
# by checking the /proc directory.
# @param pid - the pid to check for
# @return True if a process with provided pid is running, otherwise False
def isprocessrunning(pid):
    return os.path.exists("/proc/%d"%pid)
    
def run_server():
    optparser = create_optparser()
    optparser.add_option(
        "--foreground", action="store_true",
        default=False,
        help="don't detach the process"
    )
    optparser.add_option(
        "--logfile", type="path", action="store",
        help="location of the log file",
    )
    optparser.add_option(
        "--pidfile", type="path", action="store",
        default="/var/run/baltrad-bdb-server.pid",
        help="location of the pid file"
    )

    opts, args = optparser.parse_args()
    conf = read_config(opts.conffile)

    pidfile=TimeoutPIDLockFile(opts.pidfile, acquire_timeout=0)

    daemon_ctx = daemon.DaemonContext(
        working_directory="/",
        chroot_directory=None,
        stdout=sys.stdout if opts.foreground else None,
        stderr=sys.stderr if opts.foreground else None,
        detach_process=not opts.foreground,
        pidfile=pidfile
    )
    
    server_type = conf["baltrad.bdb.server.type"]
    if server_type not in ["cherrypy", "werkzeug"]:
        raise SystemExit("invalid server type in config %s" % server_type)
    
    server_uri = conf["baltrad.bdb.server.uri"]
    
    # try locking the pidfile to report possible errors to the user
    tryagain = False
    try:
        with pidfile:
            pass
    except lockfile.AlreadyLocked:
        tryagain = True
    except lockfile.LockFailed:
        tryagain = True

    if tryagain:
        pid = lockfile.pidlockfile.read_pid_from_pidfile(opts.pidfile)
        if pid != None and not isprocessrunning(pid):
            try:
                message = "pidfile exists but it seems like process is not running, probably due to an uncontrolled shutdown. Resetting.\n"
                sys.stderr.write(message)
                os.remove(opts.pidfile)
            except:
                pass
    
        try:
            with pidfile:
                pass
        except lockfile.AlreadyLocked:
            raise SystemExit("pidfile already locked: %s" % opts.pidfile)
        except lockfile.LockFailed:
            raise SystemExit("failed to lock pidfile: %s" % opts.pidfile)
    
    with daemon_ctx:
        configure_logging(opts, get_logging_level(conf))
        sys.excepthook = excepthook
        application = app.from_conf(conf)
        if server_type == "werkzeug":
            app.serve(server_uri, application)
        elif server_type == "cherrypy":
            from cherrypy import wsgiserver
            import urlparse
            parsedurl = urlparse.urlsplit(server_uri)
            cherryconf = conf.filter("baltrad.bdb.server.cherrypy.")
            nthreads = cherryconf.get_int("threads", 10)
            nbacklog = cherryconf.get_int("backlog", 5)
            ntimeout = cherryconf.get_int("timeout", 10)            
            server = wsgiserver.CherryPyWSGIServer((parsedurl.hostname, parsedurl.port), application,
                numthreads=nthreads, request_queue_size=nbacklog, timeout=ntimeout)
            try:
                server.start()
            finally:
                server.stop()

